#include "udplib.h"

udp_socket::udp_socket()
{
	m_sock			= INVALID_SOCKET;
	m_boundtopeer	= false;
	m_allow			= INADDR_ANY;

	m_lastport		= 0;
	m_lastip		= 0;
	m_connected     = false;
	m_accept_callback = 0;
}

udp_socket::~udp_socket()
{
	close();
}

bool udp_socket::create()
{
	m_sock = socket( AF_INET, SOCK_DGRAM, 0 ) ;	

	return m_sock != INVALID_SOCKET;
}

int udp_socket::send(void *buf, int num_bytes)
{
	if(m_boundtopeer)
	{
		return ::send(m_sock, (const char *)buf, num_bytes, 0);
	}
	else
	{
		return sendto(buf, num_bytes, m_lastip, m_lastport);
	}
}

int udp_socket::sendto(void *buf, int num_bytes, DWORD ip, int port)
{
	SOCKADDR_IN sa;

	sa.sin_family           = AF_INET;
	sa.sin_port             = htons (port);
	sa.sin_addr.S_un.S_addr = htonl (ip);

	return ::sendto(m_sock, (const char *)buf, num_bytes, 0, (SOCKADDR *)&sa, sizeof(sa));
}

bool udp_socket::connect(DWORD ip, int port)
{
	SOCKADDR_IN sa;

	if(m_sock == INVALID_SOCKET)
		return false;	
	
	if(ip == 0 || ip == -1)
		return false;

	sa.sin_family           = AF_INET;
	sa.sin_port             = htons (port);
	sa.sin_addr.S_un.S_addr = htonl (ip);

	if(::connect(m_sock, (SOCKADDR *)&sa, sizeof(sa)))
		return false;

	m_connected   = true;
	m_boundtopeer = true;
	return true;
}

bool udp_socket::connect(const char *hostname, int port)
{
	return connect(lookup_ip(hostname), port);
}

bool udp_socket::listen(int port, DWORD ip)
{
	SOCKADDR_IN sa;

	sa.sin_family				= AF_INET;
	sa.sin_port					= htons (port);
	sa.sin_addr.S_un.S_addr		= htonl (ip);

	if(::bind(m_sock, (SOCKADDR *)&sa, sizeof(sa)))
		return false;

	m_connected   = true;
	m_boundtopeer = false;

	return true;
}

bool udp_socket::accept(socket_base **newsock, DWORD *clientip)
{
	return false;
}

int udp_socket::select(long timeout)
{
	struct	fd_set  recv_set;
	struct	timeval tv ;
	
	tv.tv_sec  = timeout;
	tv.tv_usec = 0;
	
	FD_ZERO(&recv_set);
	FD_SET(m_sock, &recv_set);
	
	return ::select(0, &recv_set, 0, 0, &tv);
}

int udp_socket::recv(void *buf, int num_bytes, long timeout)
{
	if(timeout != INFINITE)
	{
		int	result = select(timeout);
	
		if(result == 0)
			return 0;
	
		if(result == SOCKET_ERROR)
			return SOCKET_ERROR;
	}
	
	if(m_boundtopeer)
	{
		return ::recv(m_sock, (char *)buf, num_bytes, 0);
	}
	else
	{
		DWORD ip;
		int   port;
		int   retval;

		for(;;)
		{
			if((retval = recvfrom(buf, num_bytes, &ip, &port)) <= 0)
				break;

			if(m_allow == INADDR_ANY || ip == m_allow)
			{
				m_lastip   = ip;
				m_lastport = port;
				break;
			}
			else
			{
				if(m_accept_callback)
					m_accept_callback(ip, m_accept_param);
			}
		}

		return retval;
	}
}

int udp_socket::recv_exact(void *buf, int num_bytes, long timeout)
{
	return 0;
}


int udp_socket::recvfrom(void *buf, int num_bytes, DWORD *ip, int *port, long timeout)
{
	SOCKADDR_IN sa;
	int len = sizeof(sa);
	int status;

	if(timeout != INFINITE)
	{
		int	result = select(timeout);
	
		if(result == 0)
			return 0;
	
		if(result == SOCKET_ERROR)
			return SOCKET_ERROR;
	}
	
	status = ::recvfrom(m_sock, (char *)buf, num_bytes, 0, (SOCKADDR *)&sa, &len);

	if(status >= 0)
	{
		if(ip)   *ip   = ntohl(*(DWORD *)&sa.sin_addr.S_un.S_addr);
		if(port) *port = ntohs(sa.sin_port);
	}

	return status;
}

bool udp_socket::close()
{
	SOCKET tmp_sock = m_sock;

	if(InterlockedExchange((PLONG)&m_sock, INVALID_SOCKET) != INVALID_SOCKET)
	{
		int ret = closesocket(tmp_sock);

		m_sock		  = INVALID_SOCKET;
		m_boundtopeer = false;
		m_lastport	  = 0;
		m_lastip	  = 0;
		m_allow		  = INADDR_ANY;
		m_connected   = false;

		return (ret == 0);
	}
	else
	{
		return false;
	}
}

bool udp_socket::shutdown(int how)
{
	return ::shutdown(m_sock, how) == 0 ? true : false;
}



void udp_socket::allow(DWORD ip, acceptproc ap, DWORD param)
{
	m_allow = ip;
	m_accept_callback = ap;
	m_accept_param    = param;
}

bool udp_socket::getpeer(DWORD *ip, int *port)
{
	struct sockaddr_in	sockaddr;
	int		namelen = sizeof(sockaddr);

	if(!getpeername(m_sock, (struct sockaddr *)&sockaddr, &namelen))
	{
		if(ip)   *ip   = ntohl(sockaddr.sin_addr.S_un.S_addr);
		if(port) *port = ntohs(sockaddr.sin_port);
		return true;
	}
	else
	{
		if(ip)    *ip  = m_lastip;
		if(port) *port = m_lastport;
	}
	
	return false;
}

bool udp_socket::connected()
{
	return m_connected;
}

/*char *htoa(DWORD ip)
{
	static char buf[32];

	sprintf(buf, "%d.%d.%d.%d", (ip >> 24) & 0xff, (ip >> 16) & 0xff, (ip >> 8) & 0xff, (ip >> 0) & 0xff);

	return buf;
}*/